"""Test testing_utils.py

To test:
```zsh
pytest nucleobench/common/testing_utils_test.py
```
"""
import random

from nucleobench.common import string_utils
from nucleobench.common import testing_utils


def test_dummy_inference_correctness():
    """Test that dummy net properly counts chars."""
    seq_len = 7
    batch_size = 6
    vocab = ['A', 'C', 'G', 'T']
    for vocab_i, letter_to_count in enumerate(vocab):
        seqs = []
        for _ in range(batch_size):
            seqs.append(''.join(random.choices(vocab, k=seq_len)))
        #m = testing_utils.CountLetterModel(vocab_i=vocab_i, **model_args)
        m = testing_utils.CountLetterModel(vocab_i=vocab_i, call_is_on_strings=False)
        tensors = string_utils.dna2tensor_batch(seqs, vocab_list=vocab)
        ret = m(tensors)
        for seq, count in zip(seqs, ret):
             assert seq.count(letter_to_count) == count
             

def test_tism():
    m = testing_utils.CountLetterModel(vocab_i=1)  # Count C
    _, tism = m.tism('ACTG')
    
    for base_c, cur_tism in zip('ACTG', tism):
        if base_c == 'C':
            assert cur_tism['A'] == cur_tism['T'] == cur_tism['G'] < 0
        else:
            for nt in ['A', 'T', 'G']:
                if nt == base_c: continue
                assert cur_tism[nt] == 0
            cur_tism['C'] > 0
            
def test_tism_bp_consistency():
    """Test that single BP tism."""
    m = testing_utils.CountLetterModel(vocab_i=1)  # Count C
    _, tism = m.tism('ACT')
    
    for idx in range(3):
        _, tism_cur = m.tism('ACT', [idx])
        assert tism[idx] == tism_cur[0]
    


from nucleobench.models import model_class as mc
from nucleobench.models.bpnet import model_def as bpnet_model_def
from nucleobench.models.grelu.enformer import model_def as enformer_model_def
from nucleobench.models.dummy import model_def as dummy_model_def
from nucleobench.models.malinois import model_def as malinois_model_def
from nucleobench.models.substring_count_net import model_def as substring_model_def


MODELS_ = {
    'dummy': dummy_model_def.DummyModel,
    'enformer': enformer_model_def.Enformer,
    'malinois': malinois_model_def.Malinois,
    'substring_count': substring_model_def.CountSubstringModel,
    'bpnet': bpnet_model_def.BPNet,
}

def get_model(model_name: str) -> mc.ModelClass:
    return MODELS_[model_name]


"""Utilities for parsing arguments."""

from typing import Iterable

import argparse
import dataclasses

@dataclasses.dataclass
class ParsedArgs:
    main_args: argparse.Namespace
    model_init_args: argparse.Namespace
    opt_init_args: argparse.Namespace
    opt_run_args: argparse.Namespace
    
    
def parse_long_start_sequence(known_args: argparse.Namespace) -> argparse.Namespace:
    """Parse a long start sequence from a file."""
    assert known_args.seed_sequence.startswith('local://')
    local_fileloc = known_args.seed_sequence[len('local://'):]
    with open(local_fileloc, 'r') as f:
        known_args.seed_sequence = f.read()
    return known_args


def possibly_parse_positions_to_mutate(known_args: argparse.Namespace) -> argparse.Namespace:
    """Possibly parse `positions_to_mutate` from a file, or leave it untouched, depending on the value."""
    if isinstance(known_args.positions_to_mutate, str) and known_args.positions_to_mutate.startswith('local://'):
        local_fileloc = known_args.positions_to_mutate[len('local://'):]
        with open(local_fileloc, 'r') as f:
            loc_str = f.read()
        known_args.positions_to_mutate = [int(x) for x in loc_str.split('\n')]
    elif known_args.positions_to_mutate is None or known_args.positions_to_mutate == '' or known_args.positions_to_mutate == []:
        known_args.positions_to_mutate = None
    else:
        assert isinstance(known_args.positions_to_mutate, list), (type(known_args.positions_to_mutate), known_args.positions_to_mutate)
        known_args.positions_to_mutate = [int(x) for x in known_args.positions_to_mutate.split(',')]
    return known_args


def handle_leftover_args(known_args: argparse.Namespace, leftover_args: Iterable):
    """Handle leftover arguments, either by failing or by ignoring them."""
    if known_args.ignore_empty_cmd_args:
        # Check that every "value" is either `None` or `empty`. If so, allow it to continue.
        for i in leftover_args:
            if i.startswith('--'):
                if '=' in i:
                    arg_val = i.split('=')[1]
                    if arg_val not in [None, '']:
                        raise ValueError(f'Unused arg, not empty: {leftover_args}')
                continue
            else: 
                if i not in [None, '']:
                    raise ValueError(f'Unused arg, not empty: {leftover_args}')
    else:
        raise ValueError(f'Unused args: {leftover_args}')
    

def str_to_bool(s):
    if s.lower() in ('yes', 'true', 't', '1'):
        return True
    elif s.lower() in ('no', 'false', 'f', '0'):
        return False
    else:
        raise argparse.ArgumentTypeError('Boolean value expected.')


"""Library for smoothgrad and genome-specific attribution methods.

Ref:
1. [Correcting gradient-based interpretations of deep neural networks for genomics](https://genomebiology.biomedcentral.com/articles/10.1186/s13059-023-02956-3)
2. [SmoothGrad: removing noise by adding noise](https://arxiv.org/abs/1706.03825)
3. [Quick and effective approximation of in silico saturation mutagenesis experiments with first-order taylor expansion](https://pubmed.ncbi.nlm.nih.gov/39286491/)

TODO(joelshor): Consider using public version of attribution tools, such as:
- PyTorch Smoothgrad: https://github.com/pkmr06/pytorch-smoothgrad
- PyTorch GradCam and others: https://github.com/jacobgil/pytorch-grad-cam?tab=readme-ov-file
- PyTorch Smoothgrad and others: https://tf-explain.readthedocs.io/en/latest/

To test locally:
```zsh
python -m nucleobench.common.attribution_lib
```
"""

import gc
import numpy as np
import torch
from typing import Callable, Optional


TISMOutputType = list[dict[str, float]]
SmoothgradVocabType = list[dict[str, torch.Tensor]]
TISMLocationsType = list[int]


def noise_inputs(
    input_tensor: torch.Tensor,
    noise_stdev: float,
    times: int,
    ) -> torch.Tensor:
    """Generates noisy inputs.

    NOTE: For simplicity, for now, we work with SINGLE TENSORS. Assume no batch dimension.

    Args:
        input_tensor: Input tensor. Doesn't have to be genomic. Should NOT be batched.
        noise_stdev: Noise to add.
        times: Number of times to add noise.
    """
    if noise_stdev < 0:
        raise ValueError(f'Requires non-negative noise stdev: {noise_stdev}')
    x = input_tensor  # Syntactic sugar.
    
    # Stack N versions of the input, to add uncorrelated noise to.
    with torch.no_grad():
        x = x.unsqueeze(0)
        x = x.repeat([times] + [1] * (x.ndim-1))

        # Add noise for hte smoothgrad algorithm.
        if noise_stdev > 0:
            noise_to_add = torch.normal(mean=torch.zeros(x.shape), std=noise_stdev)
            x += noise_to_add
    return x

def noisy_grads_torch(
        input_tensor: torch.Tensor, 
        model: Callable[[torch.Tensor], torch.Tensor], 
        noise_stdev: float, 
        times: int,
        idxs: Optional[TISMLocationsType] = None,
        ) -> torch.Tensor:
    """Generates noisy gradients from a function.

    NOTE: For simplicity, for now, we work with SINGLE TENSORS. Assume no batch dimension.

    This replicates the input `times` times, and runs it through the network all at once.

    TODO(joelshor): Add batching, for the situation where `times` is larger than the possible batch size
        of a single inference with a network.
    TODO(joelshor): Add ability to efficiently compute multiple inputs at once.

    Args:
        input_tensor (torch.Tensor): Input tensor. Doesn't have to be genomic. Should NOT be batched.
        model: PyTorch model to use. The model must return a scalar per batch element.
        noise_stdev: Noise to add.
        times: Number of times to add noise.
        idx: If present, only backprop through this location.
    """
    x = noise_inputs(
        input_tensor=input_tensor,
        noise_stdev=noise_stdev,
        times=times)

    # Run inference to get grads.
    if idxs is None:
        x_grad = x
        x_grad.requires_grad = True
    else:
        x, x_grad = apply_gradient_mask(x, idxs)
        
    y = model(x)
    y_sum = y.sum()
    y_sum.backward(retain_graph=False)
    noisy_grads = x_grad.grad.numpy()

    gc.collect()
    torch.cuda.empty_cache()

    assert noisy_grads.shape == x_grad.shape
    return noisy_grads


def smoothgrad_torch(
    input_tensor: torch.Tensor, 
    model: Callable[[torch.Tensor], torch.Tensor], 
    noise_stdev: float, 
    times: int,
    idxs: Optional[TISMLocationsType] = None,
    ) -> torch.Tensor:
    """Custom implementation of SmoothGrad.
    https://arxiv.org/pdf/1706.03825

    NOTE: For simplicity, for now, we work with SINGLE TENSORS. Assume no batch dimension.

    This replicates the input `times` times, and runs it through the network all at once.

    TODO(joelshor): Add batching, for the situation where `times` is larger than the possible batch size
        of a single inference with a network.
    TODO(joelshor): Add ability to efficiently compute multiple inputs at once.

    Args:
        input_tensor (torch.Tensor): Input tensor. Doesn't have to be genomic. Should NOT be batched.
        model: PyTorch model to use. The model must return a scalar per batch element.
        noise_stdev: Noise to add.
        times: Number of times to add noise.
        idx: If present, only backprop to this location.
        
    Returns:
        Per-nucleotide smoothgrad.
    """
    noisy_grads = noisy_grads_torch(
        input_tensor=input_tensor,
        model=model,
        noise_stdev=noise_stdev,
        times=times,
        idxs=idxs,
    )

    return np.mean(noisy_grads, axis=0)


# TODO(joelshor): Add `attribution_lib.py` test, taken from `malinois/model_def_test.py`.
def smoothgrad_tensor_to_dict(smooth_grad: torch.Tensor, vocab: list[str]) -> SmoothgradVocabType:
    """Map the smoothgrad indices to the vocab."""
    assert smooth_grad.ndim == 2
    assert list(smooth_grad.shape)[0] == len(vocab)
    def _to_dict(x: torch.Tensor) -> dict[str, torch.Tensor]:
        return {vocab[i]: x[i] for i in range(len(vocab))}
    return [_to_dict(x) for x in smooth_grad.T]


def smoothgrad_to_tism(sg: SmoothgradVocabType, base_seq: str) -> TISMOutputType:
    """Returns result according to Taylor in-silico mutagenesis.
    
    Paper: https://www.cell.com/iscience/fulltext/S2589-0042(24)02032-7"""
    assert len(sg) == len(base_seq)

    tism = []
    for base_nt, sg_dict in zip(base_seq, sg):
        cur_tism = {}
        for nt, sg in sg_dict.items():
            if nt == base_nt: continue
            cur_tism[nt] = float(sg - sg_dict[base_nt])
        tism.append(cur_tism)

    return tism


def apply_gradient_mask(x: torch.Tensor, idxs: TISMLocationsType) -> tuple[torch.Tensor, torch.Tensor]:
    """Applies a gradient mask to the input tensor.
    
    NOTE: Do NOT just multiply by 0. This will run out of memory in large models.

    Returns:
        Tuple of (x, masked_x), where masked_x is the input tensor with the gradient mask applied.
    """
    assert min(idxs) >= 0
    assert max(idxs) < x.shape[2]
    assert x.ndim == 3, x.shape
    
    no_gradient = x.clone().detach()
    no_gradient.requires_grad = False
    
    x_grad = x[:, :, idxs].clone().detach()
    x_grad.requires_grad = True
    x_grad_i = {idx: i for i, idx in enumerate(idxs)}
    
    # Instead of using `torch.where`, we use this method to make our gradient tensor
    # as small as possible, to preserve memory.
    tensor_slices = [x_grad[:, :, x_grad_i[i]:x_grad_i[i]+1] if i in idxs
                     else no_gradient[:, :, i:i+1]
                     for i in range(no_gradient.shape[2])]
    x = torch.concat(tensor_slices, dim=2)
    
    return x, x_grad


"""Test utils.py

To test:
pytest nucleobench/common/string_utils_test.py
"""

import numpy as np
from nucleobench.common import string_utils as utils

def test_dna2tensor2dna():
    """Test roundtrip conversion between DNA and tensor."""
    for cur_dna in [
        'ACGT',
        'ACGTACGT',
        'ACGTACGTACGT',
        'ACGTACGTACGTACGT',
        'ACGTACGTACGTACGTACGT',
        'ACGTACGTACGTACGTACGTACGT',
        'ACGTACGTACGTACGTACGTACGTACGT',
        'ACGTACGTACGTACGTACGTACGTACGTACGT',
    ]:
        cur_tensor = utils.dna2tensor(cur_dna)
        roundtrip_dna = utils.tensor2dna(cur_tensor)
        assert cur_dna == roundtrip_dna


def test_tensor2dna_batch():
    """Test roundtrip conversion between tensor and DNA, with batch util."""
    test_dnas = [
        'ACGTACGTACGT',
        'CGTACGTACGTA',
        'GTACGTACGTAA',
        'TACGTACGTCCC',
    ]
    tensors = utils.dna2tensor_batch(test_dnas)
    roundtrip_dnas = utils.tensor2dna_batch(tensors)
    for cur_dna, cur_roundtrip_dna in zip(test_dnas, roundtrip_dnas):
        assert cur_dna == cur_roundtrip_dna


def test_load_sequences():
    """Test the string codepaths."""
    # TODO(joelshor): Add tests for the file reading.
    dna_str = 'ATGCTGG'
    dna_strs = ['ATGCCCG', 'ATAA', 'ATTGC']

    assert utils.load_sequences(dna_str)[0] == dna_str
    assert utils.load_sequences(','.join(dna_strs))[0] == dna_strs


"""Parent class for optimizers."""

from nucleobench.models import model_class

from typing import Optional, Union

SequenceType = str
SamplesType = list[str]


class SequenceOptimizer(object):
    def __init__(
        self, 
        model_fn: Union[model_class.ModelClass, callable], 
        seed_sequence: SequenceType,
        positions_to_mutate: Optional[list[int]] = None,
        ):
        raise NotImplementedError("Not implemented.")

    def run(self, n_steps: int):
        raise NotImplementedError("Not implemented.")

    def get_samples(self, n_samples: int) -> SamplesType:
        raise NotImplementedError("Not implemented.")

    @staticmethod
    def init_parser():
        raise ValueError("Not implemented.")

    @staticmethod
    def run_parser():
        raise ValueError("Not implemented.")
    
    def is_finished(self) -> bool:
        raise ValueError("Not implemented.")



VOCAB = ['A','C','G','T']
SERVICE_KEY_FILE_LOCATION = ''
GCP_OUTPUT_BUCKET_NAME = ''


MPRA_UPSTREAM  = 'ACGAAAATGTTGGATGCTCATACTCGTCCTTTTTCAATATTATTGAAGCATTTATCAGGGTTACTAGTACGTCTCTCAAGGATAAGTAAGTAATATTAAGGTACGGGAGGTATTGGACAGGCCGCAATAAAATATCTTTATTTTCATTACATCTGTGTGTTGGTTTTTTGTGTGAATCGATAGTACTAACATACGCTCTCCATCAAAACAAAACGAAACAAAACAAACTAGCAAAATAGGCTGTCCCCAGTGCAAGTGCAGGTGCCAGAACATTTCTCTGGCCTAACTGGCCGCTTGACG'
MPRA_DOWNSTREAM= 'CACTGCGGCTCCTGCGATCTAACTGGCCGGTACCTGAGCTCGCTAGCCTCGAGGATATCAAGATCTGGCCTCGGCGGCCAAGCTTAGACACTAGAGGGTATATAATGGAAGCTCGACTTCCAGCTTGGCAATCCGGTACTGTTGGTAAAGCCACCATGGTGAGCAAGGGCGAGGAGCTGTTCACCGGGGTGGTGCCCATCCTGGTCGAGCTGGACGGCGACGTAAACGGCCACAAGTTCAGCGTGTCCGGCGAGGGCGAGGGCGATGCCACCTACGGCAAGCTGACCCTGAAGTTCATCT'


"""Tests for utils.

```zsh
pytest nucleobench/optimizations/utils_test.py
```
"""

import numpy as np

from nucleobench.optimizations import utils

def test_get_locations_to_edit():
    locs = utils.get_locations_to_edit(
        positions_to_mutate=[0, 1],
        random_n_loc=2,
        rng=np.random.default_rng(42),
        method='random',
    )
    assert len(locs) == 2


"""Utils for reading from and writing to GCP.

To test this with the cloud:
```zsh
python -m nucleobench.common.gcp_utils
```
"""

from typing import Any, Generator

import argparse
import copy
import os
import pickle
import subprocess
import torch
import time
from google.cloud import storage

from nucleobench.common import constants


def get_filepath(
    base_dir: str,
    opt_method: str,
    model: str,
    exp_start_time: str,
    timestamp: str,
    ) -> str:
    return os.path.join(base_dir, f'{opt_method}_{model}', exp_start_time, f'{timestamp}.pkl')

def save_proposals(
        write_dicts: list[dict],
        args: argparse.Namespace, 
        output_path: str,
        ):
    """
    Save the proposals and associated arguments to a file.

    This function saves the generated proposals and the input arguments used for their generation
    to a file. The file is named based on the current timestamp and a random tag. The saved file can
    be placed in a local directory or on Google Cloud Storage if a 'gs://' path is provided in the arguments.

    Args:
        write_dicts: A list of dictiomary of things to write.
        args: Args for the job. Best to keep them close to the output.
        output_path: Directory to write the output to, either locally or on GCP. Format of output is:
            {output_path}/{opt_method}_{model}/{exp start time}/{reults of exp N}.pkl

    Returns:
        None
    """
    def _tensor2np(obj):
        return obj.detach().clone().numpy() if isinstance(obj, torch.Tensor) else obj
    args = {k: _tensor2np(v) for k, v in vars(args).items()}
    write_dicts = [{k: _tensor2np(v) for k, v in x.items()} for x in write_dicts]
    save_dicts = copy.deepcopy(write_dicts)

    filename = get_filepath(
        base_dir=output_path,
        opt_method=args['optimization'],
        model=args['model'],
        exp_start_time=save_dicts[0]['exp_starttime_str'],
        timestamp=time.strftime("%Y%m%d_%H%M%S"),
    )
    
    # TODO(joelshor): Used to save as torch tensors. Now, saving as a pickled dictionary
    # of numpy arrays. Eventually, save as PyArrow arrays to parquet.
    if filename.startswith('gs://'):
        bucket_name = filename.split('/')[2]
        write_str_to_gcp(
            gcs_output_path=filename,
            content=pickle.dumps(save_dicts),
            binary=True,
            bucket_name=bucket_name,
        )
    else:
        if os.path.dirname(filename) != '' and os.path.dirname(filename) != '.':
            os.makedirs(os.path.dirname(filename), exist_ok=True)
        with open(filename, 'wb') as f:
            pickle.dump(save_dicts, f)
    
    print(f'Proposals deposited at:\n\t{filename}')


def get_role_client(service_json_path: str = constants.SERVICE_KEY_FILE_LOCATION):
    try:
        gcp_client = storage.Client.from_service_account_json(service_json_path)
    except FileNotFoundError as e:
        raise ValueError(
            'GCP service key not found. You are probably trying to write output to a private '
            'bucket. To write to your own bucket, change the bucket name and service key '
            f'credentials in `nucleobench/common/constants.py`: {service_json_path}') from e
    return gcp_client

def write_str_to_gcp(
        gcs_output_path: str,
        content: Any,
        binary: bool,
        bucket_name: str = constants.GCP_OUTPUT_BUCKET_NAME,
        ):
    assert gcs_output_path.startswith('gs://'), 'gcs_output_path must be a GCS path.'
    gcs_output_path = gcs_output_path[len('gs://'):]
    bucket_name, blob_fn = gcs_output_path.split('/', 1)

    # Instantiates a client.
    storage_client = get_role_client()
    bucket = storage_client.bucket(bucket_name)
    blob = bucket.blob(blob_fn)

    write_type = 'wb' if binary else 'w'
    with blob.open(write_type) as f:
        f.write(content)
        
        
def download_gcp_folder_to_local(dir: str, local_dir: str, bucket: str = constants.GCP_OUTPUT_BUCKET_NAME):
    subprocess.call(['mkdir', '-p', local_dir])
    subprocess.call(['gcloud', 'storage', 'cp', '-r', f'gs://{bucket}/{dir}', local_dir])


def list_files_recursively(local_dir: str) -> Generator[str, None, None]:
    """Recursively lists all files in a given directory."""
    for root, dirs, files in os.walk(local_dir):
        for file in files:
            yield os.path.join(root, file)


if __name__ == '__main__':
    args = {
        'fake_arg': 'fake_value',
        'fake_arg2': 'fake_value2',
        'optimization': 'dummy',
        'model': 'dummy_model',
    }
    proposal1 = {
        'string': 'ACTC',
        'energy': 1.0,
    }
    proposal2 = {
        'string': 'ACTC',
        'energy': 2.0,
    }

    write_dict = {
        'proposals': [proposal1, proposal2],
        'exp_starttime_str': '102010',
    }

    save_proposals([write_dict, write_dict], argparse.Namespace(**args), './gcp_utils_test/dummy')
    save_proposals([write_dict, write_dict], argparse.Namespace(**args), f'gs://{constants.GCP_OUTPUT_BUCKET_NAME}/gcp_utils_test/dummy')


"""Generic priority queue with supported interface."""

import dataclasses
import heapq
import bisect


@dataclasses.dataclass(order=True)
class SearchQItem:
    energy: float
    state: str
    num_edits: int


class TwoSidedPriorityQueue(object):
    """Keeps low and high items.

    Uses a simple list.

    We need access to the "best" (to expand) and the "worst" (to remove).

    Min is better for the actual problem, but provide -{energy} for the problems
    we care about, so max is better.

    TODO(joelshor): Consider using two heaps, but this would use more memory.
    """

    def __init__(self, max_items: int):
        # Max items allowed in the queue. Internally, we will at times
        # have more elements in the queue than this.
        self.max_items = max_items
        self.q  = []

    def push(self, itm: SearchQItem):
        if len(self.q) == self.max_items and itm.energy < self.q[0].energy:
            return
        elif len(self.q) < self.max_items:
            bisect.insort(self.q, itm)
        else:
            bisect.insort(self.q, itm)
            self.q.pop(0)  # Remove the "worst"


    def push_batch(self, itms: list[SearchQItem]):
        """Push a batch of items.

        TODO(joelshor): Find a more efficient way to take advantage of batching.
        """
        for itm in itms:
            self.push(itm)


    def pop(self) -> SearchQItem:
        """Pops the 'best' item."""
        return self.q.pop(-1)


    def get(self, n_samples: int) -> list[SearchQItem]:
        if n_samples > self.max_items:
            raise ValueError('Too many items requested: {n_samples} vs {self.max_items}')
        return [x.state for x in self.q[-n_samples:]]


class OneSidedPriorityQueue(object):
    """Priority queue.

    Lower energy is better, but we want to drop bad items, so
    expect inverted sign as inputs.
    """

    def __init__(self, max_items: int):
        # Max items allowed in the queue. Internally, we will at times
        # have more elements in the queue than this.
        self.max_items = max_items
        self.q  = []
        
        
    def reset_queue(self, itms: list[SearchQItem]):
        del self.q
        heapq.heapify(itms)
        self.q = itms


    def push(self, itm: SearchQItem):
        if len(self.q) == self.max_items and itm.energy < self.q[0].energy:
            return
        elif len(self.q) < self.max_items:
            heapq.heappush(self.q, itm)
        else:
            heapq.heappushpop(self.q, itm)


    def push_batch(self, itms: list[SearchQItem]):
        """Push a batch of items.

        TODO(joelshor): Find a more efficient way to take advantage of batching.
        """
        for itm in itms:
            self.push(itm)


    def get(self, n_samples: int) -> list[str]:
        if n_samples > self.max_items:
            raise ValueError('Too many items requested: {n_samples} vs {self.max_items}')
        return [x.state for x in heapq.nlargest(n_samples, self.q)]



"""Utilities for managing memory."""

import gc
import torch

def free_memory(debug: bool = False):
    collected = gc.collect()
    if debug:
        print(f'[free_memory] Collected: {collected}')
    torch.cuda.empty_cache()


"""Utils for testing."""

from typing import Optional

import torch

from nucleobench.common import constants
from nucleobench.common import attribution_lib_torch as att_lib
from nucleobench.common import string_utils


class CountLetterModel(torch.nn.Module):
    """Count number of occurances of first vocab letter."""

    def __init__(self, 
                 vocab_i: int = 1, 
                 flip_sign: bool = False, 
                 extra_channels: int = 0,
                 call_is_on_strings: bool = True,
                 add_unsqueeze_to_output: bool = False,
                 train_seq_len: int = 200,
                 ):
        super().__init__()
        self.vocab_i = vocab_i
        self.flip_sign = flip_sign
        self.extra_channels = extra_channels
        self.call_is_on_strings = call_is_on_strings
        self.add_unsqueeze_to_output = add_unsqueeze_to_output
        self.train_seq_len = train_seq_len

    def forward(self, x):
        assert x.ndim == 3
        assert x.shape[1] == 4, x.shape
        out_tensor = torch.sum(x[:, self.vocab_i, :], dim=[1])
        if self.flip_sign:
            out_tensor *= -1
        if self.extra_channels:
            out_tensor = torch.stack([out_tensor] + [torch.zeros_like(out_tensor)] * self.extra_channels).T
        if self.add_unsqueeze_to_output:
            out_tensor = torch.unsqueeze(out_tensor, dim=-1)
        return out_tensor

    def inference_on_tensor(self, x: torch.Tensor) -> torch.Tensor:
        return self.forward(x)

    def tism(self, x: str, idxs: Optional[int] = None) -> tuple[torch.Tensor, list[dict[str, torch.Tensor]]]:
        input_tensor = string_utils.dna2tensor(x, vocab_list=constants.VOCAB)
        sg_tensor = att_lib.smoothgrad_torch(
            input_tensor=input_tensor,
            model=self.inference_on_tensor,
            noise_stdev=0.1,
            times=3,
            idxs=idxs,
        )
        sg = att_lib.smoothgrad_tensor_to_dict(sg_tensor, vocab=constants.VOCAB)
        x_effective = x if idxs is None else [x[idx] for idx in idxs]
        sg = att_lib.smoothgrad_to_tism(sg, x_effective)
        y = self.inference_on_tensor(torch.unsqueeze(input_tensor, dim=0))
        return y, sg
    
    def inference_on_strings(self, seqs: list[str]) -> list[float]:
        torch_seq = string_utils.dna2tensor_batch(seqs)
        result = self.inference_on_tensor(torch_seq)
        return [float(x) for x in result]

    def __call__(self, x):
        if self.call_is_on_strings:
            return self.inference_on_strings(x)
        else:
            return self.inference_on_tensor(x)
    
    # Attributes needed for gRelu.
    @property
    def data_params(self):
        return {
            'tasks': {'name':[f'task{i}' for i in range(3)] + ['Neuron']},
            'train': {'seq_len': self.train_seq_len},
        }
    
    # Method needed for gRelu.
    def get_task_idxs(self, *args, **kwargs):
        return [0, 1, 2]



"""Parent class for models."""

from typing import Any, Callable, Optional

import torch
import numpy as np

SequenceType = str
SamplesType = list[str]

class ModelClass(object):
    
    @staticmethod
    def init_parser():
        raise ValueError('Not implemented.')
    
    @staticmethod
    def debug_init_args() -> dict[str, Any]:
        raise ValueError('Not implemented.')
    
    def __init__(self, model_fn: Callable, seed_sequence: SequenceType):
        raise NotImplementedError("Not implemented.")
    
    def __call__(self, x: SequenceType, return_debug_info: bool) -> np.ndarray:
        """Takes in a string or list of strings, returns a scalar value per string."""
        raise NotImplementedError("Not implemented.")
    
    
class TISMModelClass(ModelClass):
    """Model that supports TISM."""
    
    def tism(self, x: str, idx: Optional[int] = None) -> tuple[np.ndarray, Any]:
        """Returns inference and TISM for a single input."""
        raise NotImplementedError("Not implemented.")
    
    
class PyTorchDifferentiableModel(ModelClass):
    """Model that can produce differentiable, PyTorch tensors."""
    
    def inference_on_tensor(self, x: torch.Tensor) -> torch.Tensor:
        raise NotImplementedError("Not implemented.")


"""Test attribution_lib.py.

To test:
pytest nucleobench/common/attribution_lib_torch_test.py
"""

import pytest
import numpy as np
import torch
from torch import nn

from nucleobench.common import attribution_lib_torch
from nucleobench.common import testing_utils

class TestNeuralNetwork(nn.Module):
    def __init__(self):
        super().__init__()
        self.flatten = nn.Flatten()
        self.linear_relu_stack = nn.Sequential(
            nn.Linear(10, 3),
        )

    def forward(self, x):
        x = self.flatten(x)
        logits = self.linear_relu_stack(x)
        return logits
    

class Nonlinear(nn.Module):
    def __init__(self):
        super().__init__()
        self.flatten = nn.Flatten()
        self.linear_relu_stack = nn.Sequential(
            nn.Linear(10, 2),
            nn.Sigmoid(),
        )

    def forward(self, x):
        x = self.flatten(x)
        logits = self.linear_relu_stack(x)
        return logits


class FixedNN(nn.Module):
    def __init__(self):
        super().__init__()
        self.layer_array = np.array([1.0, 2.0])
        self.linear_layer = torch.tensor([self.layer_array], dtype=torch.float).t()

    def forward(self, x):
        logits = torch.matmul(x, self.linear_layer)
        return logits

@pytest.mark.parametrize('times', [1, 3, 5])
def test_expected_gradient(times):
    """Regardless of number of times, noisy grads should be the same for a linear model."""
    input_tensor = torch.randn(2)
    model = FixedNN()

    grads = attribution_lib_torch.noisy_grads_torch(
        input_tensor=input_tensor,
        model=model,
        noise_stdev=1.0,
        times=times)
    
    assert grads.shape == (times, 2)
    # For linear models, grads should all be the linear layer.
    for i in range(times):
        assert np.array_equal(grads[i, ...], model.layer_array), (
            grads[i, ...], model.layer_array, grads)
        
def test_callable():
    """Check that additional arguments are used."""
    input_tensor = torch.randn(2)
    model = FixedNN()

    def override_callable(x):
        return 2.0 * model(x)
    grads = attribution_lib_torch.noisy_grads_torch(
        input_tensor=input_tensor,
        model=override_callable,
        noise_stdev=1.0,
        times=3,
    )
    
    assert grads.shape == (3, 2)
    # For linear models, grads should all be the linear layer.
    for i in range(3):
        assert np.array_equal(grads[i, ...], [2.0, 4.0]), (
            grads[i, ...], [2.0, 4.0], grads)

def test_noisy_grads():
    input_tensor = torch.randn(10)
    model = TestNeuralNetwork()
    
    noisy_grads = attribution_lib_torch.noisy_grads_torch(
        input_tensor=input_tensor,
        model=model,
        noise_stdev=0.25,
        times=5)
    assert noisy_grads.shape == (5, 10)
    # For linear models, grads should all be the same.
    for i in range(5):
        assert np.array_equal(noisy_grads[i, ...], noisy_grads[0, ...])


def test_smoothgrad_torch():
    input_tensor = torch.randn(10)
    model = TestNeuralNetwork()
    
    grad = attribution_lib_torch.smoothgrad_torch(
        input_tensor=input_tensor,
        model=model,
        noise_stdev=0.25,
        times=5)

    assert grad.shape == (10,)


def test_noisy_grads_different():
    """In a nonlinear network, gradients should be different."""
    input_tensor = torch.randn(10)
    model = Nonlinear()
    
    noisy_grads = attribution_lib_torch.noisy_grads_torch(
        input_tensor=input_tensor,
        model=model,
        noise_stdev=1.0,
        times=5)
    assert noisy_grads.shape == (5, 10)
    # For linear models, grads should all be the same.
    for i in range(1, 5):
        assert not np.array_equal(noisy_grads[i, ...], noisy_grads[0, ...])
        
# TODO(joelshor): Add a unit test for nucleotide-specific TISM.
def test_smoothgrad_torch_idx_sanity():
    input_tensor = torch.randn(4, 5)
    model = testing_utils.CountLetterModel()
    
    grad = attribution_lib_torch.smoothgrad_torch(
        input_tensor=input_tensor,
        model=model.inference_on_tensor,
        noise_stdev=0.0,
        times=1)
    assert grad.shape == (4, 5)
    
    for idx in range(5):
        grad_singlebp = attribution_lib_torch.smoothgrad_torch(
            input_tensor=input_tensor,
            model=model.inference_on_tensor,
            noise_stdev=0.25,
            times=5,
            idxs=[idx])
        assert grad_singlebp.shape == (4, 1)
        assert np.all(grad[:, idx] == grad_singlebp[:, 0])
        
        
def test_apply_gradient_mask():
    """[idx] and idx should give the same result."""
    input_tensor = torch.randn(1, 10)
    model = TestNeuralNetwork()
    output_tensor = model(input_tensor)
    output_tensor = output_tensor.reshape(1, 1, 3)
    
    x1, x_grad = attribution_lib_torch.apply_gradient_mask(
        output_tensor, [0])
    assert x1.shape == (1, 1, 3)
    assert x_grad.shape == (1, 1, 1)
    
    x2, x_grad = attribution_lib_torch.apply_gradient_mask(
        output_tensor, [0, 1])
    assert x2.shape == (1, 1, 3)
    assert x_grad.shape == (1, 1, 2)
    
    assert (x1 == x2).all(), x1 == x2


"""Test some stuff.

To test:
```zsh
pytest nucleobench/common/priority_queue_test.py
```
"""

from nucleobench.common import priority_queue

def test_onesided_priority_queue():
    q = priority_queue.OneSidedPriorityQueue(max_items=4)
    
    def _w(s: str):
        return priority_queue.SearchQItem(
            state=s, 
            energy=int(s),
            num_edits=0,
        )
    
    for i in range(10):
        q.push(_w(str(i)))
    
    assert len(q.q) == 4
    assert set(q.get(4)) == set(['6', '7', '8', '9'])
    assert set(q.get(3)) == set(['7', '8', '9'])
    
    q.push_batch([_w('-1'), _w('10'), _w('11')])
    assert len(q.q) == 4
    assert set(q.get(4)) == set(['8', '9', '10', '11'])
    
    
def test_twosided_priority_queue():
    q = priority_queue.TwoSidedPriorityQueue(
        max_items=4,
    )
    
    def _w(s: str):
        return priority_queue.SearchQItem(
            state=s, 
            energy=int(s),
            num_edits=0,
        )
    
    for i in range(10):
        q.push(_w(str(i)))
    
    assert len(q.q) == 4
    assert q.get(4) == ['6', '7', '8', '9']
    
    q.push_batch([_w('-1'), _w('10'), _w('11')])
    assert len(q.q) == 4
    assert q.get(4) == ['8', '9', '10', '11']
    assert q.pop().state == '11'
    assert q.get(4) == ['8', '9', '10']


"""Utils for manipulating stringsl."""

import os
import numpy as np
import torch
from typing import Union
import subprocess

from nucleobench.common import constants


def dna2tensor(sequence_str: str, vocab_list=constants.VOCAB) -> torch.Tensor:
    """
    Convert a DNA sequence to a one-hot encoded tensor.

    Args:
        sequence_str (str): DNA sequence string.
        vocab_list (list): List of DNA nucleotide characters.

    Returns:
        torch.Tensor: One-hot encoded tensor representation of the sequence.
    """
    seq_tensor = np.zeros((len(vocab_list), len(sequence_str)))
    for letterIdx, letter in enumerate(sequence_str):
        seq_tensor[vocab_list.index(letter), letterIdx] = 1
    seq_tensor = torch.Tensor(seq_tensor)
    return seq_tensor


def dna2tensor_batch(
    sequence_strs: list[str], vocab_list=constants.VOCAB
) -> torch.Tensor:
    """
    Convert a DNA sequence to a one-hot encoded tensor.

    Args:
        sequence_str (str): DNA sequence string.
        vocab_list (list): List of DNA nucleotide characters.

    Returns:
        torch.Tensor: One-hot encoded tensor representation of the sequence.
    """
    seq_tensors = [dna2tensor(x, vocab_list) for x in sequence_strs]
    return torch.stack(seq_tensors)


def tensor2dna(
    tensor: Union[torch.Tensor, np.ndarray], vocab_list=constants.VOCAB
) -> str:
    """
    Convert a one-hot encoded tensor to a DNA sequence.

    Args:
        tensor: One-hot encoded Tensor or array.
        vocab_list (list): List of DNA nucleotide characters.

    Returns:
        torch.Tensor: One-hot encoded tensor representation of the sequence.
    """
    if isinstance(tensor, torch.Tensor):
        tensor = tensor.numpy()

    if tensor.ndim != 2:
        raise ValueError(f"Expected dim 2: {tensor.ndim}")
    if tensor.shape[0] != len(vocab_list):
        raise ValueError(f"Expected vocab dim to be {len(vocab_list)}): {tensor.shape}")
    if not np.all(np.isin(tensor, [0, 1])):
        raise ValueError(f"Expected values to be 0 or 1: {tensor}")
    col_sum = np.sum(tensor, axis=0)
    if not np.all(col_sum == np.ones_like(col_sum)):
        raise ValueError(f"Not onehot: {tensor}")

    # Convert one-hot tensors to string sequence.
    seq = []
    for idx in range(tensor.shape[1]):
        cur_onehot = tensor[:, idx]
        cur_char = vocab_list[np.nonzero(cur_onehot)[0][0]]
        seq.append(cur_char)
    return "".join(seq)


def tensor2dna_batch(
    tensor: Union[torch.Tensor, np.ndarray], vocab_list=constants.VOCAB
) -> list[str]:
    """
    Convert a one-hot encoded tensor to a DNA sequence.

    Args:
        tensor: One-hot encoded Tensor or array.
        vocab_list (list): List of DNA nucleotide characters.

    Returns:
        torch.Tensor: One-hot encoded tensor representation of the sequence.
    """
    if isinstance(tensor, torch.Tensor):
        tensor = tensor.numpy()
    if tensor.ndim != 3:
        raise ValueError(f"Expected dim 3: {tensor.ndim}")

    return [tensor2dna(t, vocab_list) for t in tensor]


def str2np(sequence_str: str, vocab_list=constants.VOCAB) -> np.ndarray:
    """
    Convert a DNA sequence string to a numpy array.

    Args:
        sequence_str (str): DNA sequence string. Each character is a nucleotide (e.g. A, C, T, G).
        vocab_list (list): List of DNA nucleotide characters.

    Returns:
        np.ndarray: DNA sequence array. Each array element is a nucleotide index (e.g. 0, 1, 2, 3).
    """
    return np.array([vocab_list.index(letter) for letter in sequence_str])


def np2str(sequence_np: np.ndarray, vocab_list=constants.VOCAB) -> str:
    """
    Convert a DNA sequence array to a string.

    Args:
        sequence_np (np.ndarray): DNA sequence array. Each array element is a nucleotide index (e.g. 0, 1, 2, 3).
        vocab_list (list): List of DNA nucleotide characters.

    Returns:
        str: DNA sequence string. Each character is a nucleotide (e.g. A, C, T, G).
    """
    return "".join([vocab_list[letter] for letter in sequence_np])


SeqOrSeqsType = Union[str, list[str]]


def load_sequences(
    artifact_path_or_seq: str, download_path: str = "./"
) -> tuple[SeqOrSeqsType, str]:
    """Load start sequences from file or gcs. Can be a single sequence or a list of sequences.

    Use the same download style as other places in this file.

    Input can either be:
        - a local file path
        - a gcs file path
        - a single sequence
        - a comma delimited list of sequences

    Returns:
        (sequence or sequences, a comment on where the sequence came from.)
    """
    ret_comment = None
    if not set(artifact_path_or_seq).issubset(set(constants.VOCAB + [","])):
        # Assume it's a path.
        artifact_path = artifact_path_or_seq
        ret_comment = artifact_path
        if artifact_path.startswith("gs://"):
            # TODO(joelshor): Read using google.storage, not subprocess.
            subprocess.call(["gsutil", "cp", artifact_path, download_path])
            artifact_path = os.path.join(download_path, os.path.basename(artifact_path))
        with open(artifact_path, "r") as f:
            seq_or_seqs = f.read().strip()
    else:
        seq_or_seqs = artifact_path_or_seq
        if "," in seq_or_seqs:
            ret_comment = "From command line, comma delimited."
        else:
            ret_comment = "From command line."

    # Determine whether it's a single sequence or a list of sequences.
    if "," in seq_or_seqs:
        seq_or_seqs = seq_or_seqs.split(",")

    assert ret_comment is not None
    return seq_or_seqs, ret_comment



from nucleobench.optimizations import optimization_class as oc

from nucleobench.optimizations.beam_search import beam_ordered, beam_unordered
from nucleobench.optimizations.dummy import random_perturbation
from nucleobench.optimizations.fastseqprop_torch import fs
from nucleobench.optimizations.ledidi import ledidi
from nucleobench.optimizations.directed_evolution import directed_evolution
from nucleobench.optimizations.ada.adalead import adalead_ref
from nucleobench.optimizations.simulated_annealing import simulated_annealing

OPTIMIZATIONS_REQUIRING_TISM_ = {
    "beam_search": beam_ordered.OrderedBeamSearch,
    "beam_search_unordered": beam_unordered.UnorderedBeamSearch,
}

OPTIMIZATIONS_REQUIRING_PYTORCH_DIFF_ = {
    "fastseqprop": fs.FastSeqProp,
    "ledidi": ledidi.Ledidi,
}

GENERAL_OPTIMIZATIONS_ = {
    "adalead": adalead_ref.AdaLeadRef,
    "directed_evolution": directed_evolution.DirectedGreedEvolution,
    "dummy": random_perturbation.RandomPerturbation,
    "simulated_annealing": simulated_annealing.SimulatedAnnealing,
}

OPTIMIZATIONS_ = {}
OPTIMIZATIONS_.update(OPTIMIZATIONS_REQUIRING_TISM_)
OPTIMIZATIONS_.update(OPTIMIZATIONS_REQUIRING_PYTORCH_DIFF_)
OPTIMIZATIONS_.update(GENERAL_OPTIMIZATIONS_)


def get_optimization(opt_name: str) -> oc.SequenceOptimizer:
    return OPTIMIZATIONS_[opt_name]



"""Common utils for optimization algorithms."""

from typing import Generator, Optional

import numpy as np
import random


def get_locations_to_edit(
    positions_to_mutate: list[int],
    random_n_loc: int,
    rng: np.random.Generator,
    method: str,
) -> list[int]:
    """Selects locations to edit."""
    assert random_n_loc > 0
    assert random_n_loc <= len(positions_to_mutate)
        
    if method == 'all':
        return positions_to_mutate
    elif method == 'random':
        return rng.choice(positions_to_mutate, size=random_n_loc, replace=False)
    else:
        raise ValueError('Arg not recognized.')
    
    
def generate_single_edit_mutants(
    base_str: str, 
    loc_to_edit: int,
    alphabet: list[str],
    rng: np.random.Generator,
    method: str,
    ) -> Generator[str, None, None]:
    """Return a generator of potential next strings."""
    assert isinstance(alphabet, list)
    assert len(alphabet) > 1
    
    for loc in loc_to_edit:
        if method == 'all':
            chars = alphabet
        elif method == 'random':
            chars = rng.choice(alphabet)
        else:
            raise ValueError('Arg not recognized.')
        for c in chars:
            yield base_str[:loc] + c + base_str[loc + 1:]
            
            
def generate_single_mutant_multiedits(
    base_str: str, 
    locs_to_edit: list[int],
    alphabet: list[str],
    rng: np.random.Generator,
    ) -> str:
    """Return a mutant."""
    assert isinstance(alphabet, list)
    assert len(alphabet) > 1
    mutant = list(base_str)

    for i in locs_to_edit:
        # TODO(joelshor): This should be `rng.choice(set(alphabet) - mutant[i])`, 
        # but we want to keep it for consistency with the publication.
        # Expect this behavior to change in the future.
        mutant[i] = rng.choice(alphabet)
    return "".join(mutant)